/*
** Copyright (c) 2012, Romain Dura romain@shazbits.com
** 
** Permission to use, copy, modify, and/or distribute this software for any 
** purpose with or without fee is hereby granted, provided that the above 
** copyright notice and this permission notice appear in all copies.
** 
** THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES 
** WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF 
** MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY 
** SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
** WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN 
** ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR 
** IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
** Photoshop & misc math
** Blending modes, RGB/HSL/Contrast/Desaturate, levels control
**
** Romain Dura | Romz
** Blog: http://mouaif.wordpress.com
** Post: http://mouaif.wordpress.com/?p=94
*/

precision highp float;

/*
** Desaturation
*/

vec4 Desaturate(vec3 color, float Desaturation)
{
	vec3 grayXfer = vec3(0.3, 0.59, 0.11);
	vec3 gray = vec3(dot(grayXfer, color));
	return vec4(mix(color, gray, Desaturation), 1.0);
}


/*
** Hue, saturation, luminance
*/

vec3 RGBToHSL(vec3 color)
{
	vec3 hsl; // init to 0 to avoid warnings ? (and reverse if + remove first part)
	
	float fmin = min(min(color.r, color.g), color.b);    //Min. value of RGB
	float fmax = max(max(color.r, color.g), color.b);    //Max. value of RGB
	float delta = fmax - fmin;             //Delta RGB value

	hsl.z = (fmax + fmin) / 2.0; // Luminance

	if (delta == 0.0)		//This is a gray, no chroma...
	{
		hsl.x = 0.0;	// Hue
		hsl.y = 0.0;	// Saturation
	}
	else                                    //Chromatic data...
	{
		if (hsl.z < 0.5)
			hsl.y = delta / (fmax + fmin); // Saturation
		else
			hsl.y = delta / (2.0 - fmax - fmin); // Saturation
		
		float deltaR = (((fmax - color.r) / 6.0) + (delta / 2.0)) / delta;
		float deltaG = (((fmax - color.g) / 6.0) + (delta / 2.0)) / delta;
		float deltaB = (((fmax - color.b) / 6.0) + (delta / 2.0)) / delta;

		if (color.r == fmax )
			hsl.x = deltaB - deltaG; // Hue
		else if (color.g == fmax)
			hsl.x = (1.0 / 3.0) + deltaR - deltaB; // Hue
		else if (color.b == fmax)
			hsl.x = (2.0 / 3.0) + deltaG - deltaR; // Hue

		if (hsl.x < 0.0)
			hsl.x += 1.0; // Hue
		else if (hsl.x > 1.0)
			hsl.x -= 1.0; // Hue
	}

	return hsl;
}

float HueToRGB(float f1, float f2, float hue)
{
	if (hue < 0.0)
		hue += 1.0;
	else if (hue > 1.0)
		hue -= 1.0;
	float res;
	if ((6.0 * hue) < 1.0)
		res = f1 + (f2 - f1) * 6.0 * hue;
	else if ((2.0 * hue) < 1.0)
		res = f2;
	else if ((3.0 * hue) < 2.0)
		res = f1 + (f2 - f1) * ((2.0 / 3.0) - hue) * 6.0;
	else
		res = f1;
	return res;
}

vec3 HSLToRGB(vec3 hsl)
{
	vec3 rgb;
	
	if (hsl.y == 0.0)
		rgb = vec3(hsl.z); // Luminance
	else
	{
		float f2;
		
		if (hsl.z < 0.5)
			f2 = hsl.z * (1.0 + hsl.y);
		else
			f2 = (hsl.z + hsl.y) - (hsl.y * hsl.z);
			
		float f1 = 2.0 * hsl.z - f2;
		
		rgb.r = HueToRGB(f1, f2, hsl.x + (1.0/3.0));
		rgb.g = HueToRGB(f1, f2, hsl.x);
		rgb.b= HueToRGB(f1, f2, hsl.x - (1.0/3.0));
	}
	
	return rgb;
}


/*
** Contrast, saturation, brightness
** Code of this function is from TGM's shader pack
** http://irrlicht.sourceforge.net/phpBB2/viewtopic.php?t=21057
*/

// For all settings: 1.0 = 100% 0.5=50% 1.5 = 150%
vec3 ContrastSaturationBrightness(vec3 color, float brt, float sat, float con)
{
	// Increase or decrease theese values to adjust r, g and b color channels seperately
	const float AvgLumR = 0.5;
	const float AvgLumG = 0.5;
	const float AvgLumB = 0.5;
	
	const vec3 LumCoeff = vec3(0.2125, 0.7154, 0.0721);
	
	vec3 AvgLumin = vec3(AvgLumR, AvgLumG, AvgLumB);
	vec3 brtColor = color * brt;
	vec3 intensity = vec3(dot(brtColor, LumCoeff));
	vec3 satColor = mix(intensity, brtColor, sat);
	vec3 conColor = mix(AvgLumin, satColor, con);
	return conColor;
}


/*
** Float blending modes
** Adapted from here: http://www.nathanm.com/photoshop-blending-math/
** But I modified the HardMix (wrong condition), Overlay, SoftLight, ColorDodge, ColorBurn, VividLight, PinLight (inverted layers) ones to have correct results
*/

#define BlendLinearDodgef 			BlendAddf
#define BlendLinearBurnf 			BlendSubstractf
#define BlendAddf(base, blend) 		(base+blend)
#define BlendSubstractf(base, blend) 	max(base + blend - 1.0, 0.0)
#define BlendLightenf(base, blend) 		max(blend, base)
#define BlendDarkenf(base, blend) 		min(blend, base)
#define BlendLinearLightf(base, blend) 	(blend < 0.5 ? BlendLinearBurnf(base, (2.0 * blend)) : BlendLinearDodgef(base, (2.0 * (blend - 0.5))))
#define BlendScreenf(base, blend) 		(1.0 - ((1.0 - base) * (1.0 - blend)))
#define BlendOverlayf(base, blend) 	(base < 0.5 ? (2.0 * base * blend) : (1.0 - 2.0 * (1.0 - base) * (1.0 - blend)))
#define BlendSoftLightf(base, blend) 	((blend < 0.5) ? (2.0 * base * blend + base * base * (1.0 - 2.0 * blend)) : (sqrt(base) * (2.0 * blend - 1.0) + 2.0 * base * (1.0 - blend)))
#define BlendColorDodgef(base, blend) 	((blend == 1.0) ? blend : min(base / (1.0 - blend), 1.0))
#define BlendColorBurnf(base, blend) 	((blend == 0.0) ? blend : max((1.0 - ((1.0 - base) / blend)), 0.0))
#define BlendVividLightf(base, blend) 	((blend < 0.5) ? BlendColorBurnf(base, (2.0 * blend)) : BlendColorDodgef(base, (2.0 * (blend - 0.5))))
#define BlendPinLightf(base, blend) 	((blend < 0.5) ? BlendDarkenf(base, (2.0 * blend)) : BlendLightenf(base, (2.0 *(blend - 0.5))))
#define BlendHardMixf(base, blend) 	((BlendVividLightf(base, blend) < 0.5) ? 0.0 : 1.0)
#define BlendReflectf(base, blend) 		((blend == 1.0) ? blend : min(base * base / (1.0 - blend), 1.0))


/*
** Vector3 blending modes
*/

// Component wise blending
#define Blend(base, blend, funcf) 		vec3(funcf(base.r, blend.r), funcf(base.g, blend.g), funcf(base.b, blend.b))

#define BlendNormal(base, blend) 		(blend)
#define BlendLighten				BlendLightenf
#define BlendDarken				BlendDarkenf
#define BlendMultiply(base, blend) 		(base * blend)
#define BlendAverage(base, blend) 		((base + blend) / 2.0)
#define BlendAdd(base, blend)           (base+blend)
#define BlendSubstract(base, blend) 	max(base + blend - vec3(1.0), vec3(0.0))
#define BlendDifference(base, blend) 	abs(base - blend)
#define BlendNegation(base, blend) 	(vec3(1.0) - abs(vec3(1.0) - base - blend))
#define BlendExclusion(base, blend) 	(base + blend - 2.0 * base * blend)
#define BlendScreen(base, blend) 		Blend(base, blend, BlendScreenf)
#define BlendOverlay(base, blend) 		Blend(base, blend, BlendOverlayf)
#define BlendSoftLight(base, blend) 	Blend(base, blend, BlendSoftLightf)
#define BlendHardLight(base, blend) 	BlendOverlay(blend, base)
#define BlendColorDodge(base, blend) 	Blend(base, blend, BlendColorDodgef)
#define BlendColorBurn(base, blend) 	Blend(base, blend, BlendColorBurnf)
#define BlendLinearDodge			BlendAdd
#define BlendLinearBurn			BlendSubstract
// Linear Light is another contrast-increasing mode
// If the blend color is darker than midgray, Linear Light darkens the image by decreasing the brightness. If the blend color is lighter than midgray, the result is a brighter image due to increased brightness.
#define BlendLinearLight(base, blend) 	Blend(base, blend, BlendLinearLightf)
#define BlendVividLight(base, blend) 	Blend(base, blend, BlendVividLightf)
#define BlendPinLight(base, blend) 		Blend(base, blend, BlendPinLightf)
#define BlendHardMix(base, blend) 		Blend(base, blend, BlendHardMixf)
#define BlendReflect(base, blend) 		Blend(base, blend, BlendReflectf)
#define BlendGlow(base, blend) 		BlendReflect(blend, base)
#define BlendPhoenix(base, blend) 		(min(base, blend) - max(base, blend) + vec3(1.0))
#define BlendOpacity(base, blend, F, O) 	(F(base, blend) * O + blend * (1.0 - O))


// Hue Blend mode creates the result color by combining the luminance and saturation of the base color with the hue of the blend color.
vec3 BlendHue(vec3 base, vec3 blend)
{
	vec3 baseHSL = RGBToHSL(base);
	return HSLToRGB(vec3(RGBToHSL(blend).r, baseHSL.g, baseHSL.b));
}

// Saturation Blend mode creates the result color by combining the luminance and hue of the base color with the saturation of the blend color.
vec3 BlendSaturation(vec3 base, vec3 blend)
{
	vec3 baseHSL = RGBToHSL(base);
	return HSLToRGB(vec3(baseHSL.r, RGBToHSL(blend).g, baseHSL.b));
}

// Color Mode keeps the brightness of the base color and applies both the hue and saturation of the blend color.
vec3 BlendColor(vec3 base, vec3 blend)
{
	vec3 blendHSL = RGBToHSL(blend);
	return HSLToRGB(vec3(blendHSL.r, blendHSL.g, RGBToHSL(base).b));
}

// Luminosity Blend mode creates the result color by combining the hue and saturation of the base color with the luminance of the blend color.
vec3 BlendLuminosity(vec3 base, vec3 blend)
{
	vec3 baseHSL = RGBToHSL(base);
	return HSLToRGB(vec3(baseHSL.r, baseHSL.g, RGBToHSL(blend).b));
}


/*
** Gamma correction
** Details: http://blog.mouaif.org/2009/01/22/photoshop-gamma-correction-shader/
*/

#define GammaCorrection(color, gamma)								pow(color, 1.0 / gamma)

/*
** Levels control (input (+gamma), output)
** Details: http://blog.mouaif.org/2009/01/28/levels-control-shader/
*/

#define LevelsControlInputRange(color, minInput, maxInput)				min(max(color - vec3(minInput), vec3(0.0)) / (vec3(maxInput) - vec3(minInput)), vec3(1.0))
#define LevelsControlInput(color, minInput, gamma, maxInput)				GammaCorrection(LevelsControlInputRange(color, minInput, maxInput), gamma)
#define LevelsControlOutputRange(color, minOutput, maxOutput) 			mix(vec3(minOutput), vec3(maxOutput), color)
#define LevelsControl(color, minInput, gamma, maxInput, minOutput, maxOutput) 	LevelsControlOutputRange(LevelsControlInput(color, minInput, gamma, maxInput), minOutput, maxOutput)

varying vec2 layer1_coordinates;
varying vec2 base_coordinates;
uniform lowp sampler2D base_texture;
uniform lowp sampler2D layer1_texture;
uniform int blend_mode;
uniform float layer1_contrast;
uniform float layer1_brightness;
uniform float layer1_saturation;


// averaged pixel intensity from 3 color channels
float avg_intensity(in vec4 pix) {
 return (pix.r + pix.g + pix.b)/3.;
}

vec4 get_pixel(in vec2 coords, in float dx, in float dy) {
 return texture2D(layer1_texture,coords + vec2(dx, dy));
}


vec4 convolute(in vec2 coords/*, in float matrix[9]*/) {
	float dxtex = 1.0 / 480.0 /*image width*/;
	float dytex = 1.0 / 640.0 /*image height*/;
	vec4 pix[9];
	int k = -1;
	float delta;
	vec4 sum = vec4(0.0);
	  // read neighboring pixel intensities
	for (int i=-1; i<2; i++) {
		for(int j=-1; j<2; j++) {
	   		k++;
			pix[k] = get_pixel(coords,float(i)*dxtex, float(j)*dytex);
			if (i == 0 && j == 0) {
				sum += 8.0* pix[k];
			} else {
				sum -= pix[k];
			}
		}
	}
	
	return vec4(sum.rgb, 1.0);
}

vec4 sharpen(in vec2 coords) {
	float matrix[9];
	matrix[1] = matrix[3] = matrix[5] = matrix[7] = -1.0;
	matrix[0] = matrix[2] = matrix[6] = matrix[8] = 0.0;
	matrix[4] = 5.0;
	return convolute(coords/*, matrix*/);
}

vec4 detectEdges(in vec2 coords) {
	float matrix[9];
	matrix[0] = matrix[1] = matrix[2] = matrix[3] = matrix[5] = matrix[6] = matrix[7] = matrix[8] = -1.0;
	matrix[4] = 8.0;
	return convolute(coords/*, matrix*/) + 0.51;
}

/* Angle of halftone grid (degrees; positive = counterclockwise) */
const float angle = -45.0;

/* Contrast control */
const float contrastDelta = 0.3; // higher -> grey gets darker
const float brightness = 0.0; // analog for white
const float blackness = 1.1; // higher -> larger areas completely covered by dots

/* smoothness black to white (pseudo anti-aliasing). */
const float smooth = 0.2;

const vec4 dotColor = vec4(0.0, 0.0, 0.0, 1.0);
const vec4 backgroundColor = vec4(1.0, 1.0, 1.0, 0.0);
const ivec2 halftoneResolution = ivec2(150, 150);

vec4 halftone(in vec2 coordinates)
{
	const mat2 rotate = mat2(cos(angle), -sin(angle),
	                         sin(angle), cos(angle));
	const mat2 inverse_rotate = mat2(cos(angle), sin(angle),
	                                -sin(angle), cos(angle));

	/* Distance to next dot divided by two. */ 
	const vec2 halfLineDist = vec2(1.0)/vec2(halftoneResolution)/vec2(2.0);

	/* Find center of the halftone dot. */
	vec2 center =  rotate * coordinates;
	center = floor(center * vec2(halftoneResolution)) / vec2(halftoneResolution);
	center += halfLineDist;
	center = inverse_rotate * center;

	/* Only red (texture is grey scale) */
	float luminance = texture2D(layer1_texture, center).r;


	/* Radius of the halftone dot. */
	float radius = sqrt(2.0)*halfLineDist.x*(1.0 - luminance)*blackness;

	float contrast = 1.0 + (contrastDelta)/(2.0);
	float radiusSqrd = contrast * pow(radius,2.0)
		- (contrastDelta * halfLineDist.x*halfLineDist.x)/2.0
		- brightness * halfLineDist.x*halfLineDist.x;


	vec2 power = pow(abs(center-coordinates),vec2(2.0));
	float pixelDist2 = power.x + power.y; // Distance pixel to center squared.

	float delta = smooth*pow(halfLineDist.x,2.0);
	float gradient = smoothstep(radiusSqrd-delta, radiusSqrd+delta, pixelDist2);
	return mix(dotColor, backgroundColor, gradient);
}

/*
enum FATEXTURE2D_BLEND_ENUM {
    kFATexture2DBlendNormal = 0,
    kFATexture2DBlendDarken = 1,
    kFATexture2DBlendMultiply = 2,
    kFATexture2DBlendColorBurn = 3,
    kFATexture2DBlendLinearBurn = 4,
    kFATexture2DBlendLighten = 5,
    kFATexture2DBlendScreen = 6,
    kFATexture2DBlendColorDodge = 7,
    kFATexture2DBlendLinearDodge = 8,
    kFATexture2DBlendOverlay = 9,
    kFATexture2DBlendSoftLight = 10,
    kFATexture2DBlendHardLight = 11,
    kFATexture2DBlendVividLight = 12,
    kFATexture2DBlendLinearLight = 13,
    kFATexture2DBlendPinLight = 14,
    kFATexture2DBlendHardMix = 15,
    kFATexture2DBlendDifference = 16,
    kFATexture2DBlendExclusion = 17,
    kFATexture2DBlendSubstract = 18,
    kFATexture2DBlendHue = 19,
    kFATexture2DBlendSaturation = 20,
    kFATexture2DBlendColor = 21,
    kFATexture2DBlendLuminosity = 22,
    kFATexture2DBlendNegation = 23,
    kFATexture2DBlendAverage = 24,
    kFATexture2DBlendReflect = 25,
    kFATexture2DBlendGlow = 26,
    kFATexture2DBlendPhoenix = 27
};
*/

void main(void) {
	lowp vec4 base_pixel = texture2D(base_texture, base_coordinates);
	lowp vec4 layer1_pixel = texture2D(layer1_texture, layer1_coordinates);
        
	lowp vec3 processing_pixel = layer1_pixel.rgb;
    lowp float alpha_factor = layer1_pixel.a;
    
	if (blend_mode == 0) {
		processing_pixel = layer1_pixel.rgb;
        alpha_factor = 1.0;
	} else if (blend_mode == 1) {
		processing_pixel = BlendDarken(base_pixel.rgb, layer1_pixel.rgb);
        alpha_factor = 1.0;
	} else if (blend_mode == 2) {
		processing_pixel = BlendMultiply(base_pixel.rgb, layer1_pixel.rgb);
        alpha_factor = 1.0;
	} else if (blend_mode == 3) {
		processing_pixel = BlendColorBurn(base_pixel.rgb, layer1_pixel.rgb);
        alpha_factor = 1.0;
	} else if (blend_mode == 4) {
		processing_pixel = BlendLinearBurn(base_pixel.rgb, layer1_pixel.rgb);
        alpha_factor = 1.0;
	} else if (blend_mode == 5) {
		processing_pixel = BlendLighten(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 6) {
		processing_pixel = BlendScreen(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 7) {
		processing_pixel = BlendColorDodge(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 8) {
		processing_pixel = BlendLinearDodge(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 9) {
		processing_pixel = BlendOverlay(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 10) {
		processing_pixel = BlendSoftLight(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 11) {
		processing_pixel = BlendHardLight(base_pixel.rgb, layer1_pixel.rgb);
        alpha_factor = 1.0;
	} else if (blend_mode == 12) {
		processing_pixel = BlendVividLight(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 13) {
		processing_pixel = BlendLinearLight(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 14) {
		processing_pixel = BlendPinLight(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 15) {
		processing_pixel = BlendHardMix(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 16) {
		processing_pixel = BlendDifference(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 17) {
		processing_pixel = BlendExclusion(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 18) {
		processing_pixel = BlendSubstract(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 19) {
		processing_pixel = BlendHue(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 20) {
		processing_pixel = BlendSaturation(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 21) {
		processing_pixel = BlendColor(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 22) {
		processing_pixel = BlendLuminosity(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 23) {
		processing_pixel = BlendNegation(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 24) {
		processing_pixel = BlendAverage(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 25) {
		processing_pixel = BlendReflect(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 26) {
		processing_pixel = BlendGlow(base_pixel.rgb, layer1_pixel.rgb);
	} else if (blend_mode == 27) {
		processing_pixel = BlendPhoenix(base_pixel.rgb, layer1_pixel.rgb);
	}

	processing_pixel = ContrastSaturationBrightness(processing_pixel, layer1_brightness, layer1_saturation, layer1_contrast);
    
    lowp vec4 processed_pixel = vec4(BlendAdd((1.0-layer1_pixel.a)*base_pixel.rgb, alpha_factor*processing_pixel.rgb),base_pixel.a*(1.0-layer1_pixel.a)+base_pixel.a*layer1_pixel.a);
    
    
    //mediump vec4 processed_pixel = vec4(processing_pixel.rgb, base_pixel.a);
	gl_FragColor = processed_pixel;
}
